#: Conversion factor from Gaussian units to SI units
SI = 2.3070775523417355e-28

#: Conversion factor from Gaussian units to electron volts / Angstroms
eV_A = 14.399645478425667

#: Conversion factor from Gaussian units to kilocalories per mole / Angstroms
kcalmol_A = 332.0637132991921

#: Conversion factor from Gaussian units to kilojoules per mole / Angstroms
kJmol = 1389.3545764438197
